@@ -3,7 +3,9 @@ package manifests
 import (
 	"context"
 	"encoding/json"
+	"errors"
 	stderrors "errors"
+	"fmt"
 	"io"
 
 	"github.com/containers/common/pkg/manifests"
@@ -21,12 +23,17 @@ import (
 	"github.com/containers/storage/pkg/lockfile"
 	digest "github.com/opencontainers/go-digest"
 	v1 "github.com/opencontainers/image-spec/specs-go/v1"
-	"github.com/pkg/errors"
 	"github.com/sirupsen/logrus"
 )
 
 const instancesData = "instances.json"
 
+// LookupReferenceFunc return an image reference based on the specified one.
+// The returned reference can return custom ImageSource or ImageDestination
+// objects which intercept or filter blobs, manifests, and signatures as
+// they are read and written.
+type LookupReferenceFunc func(ref types.ImageReference) (types.ImageReference, error)
+
 // ErrListImageUnknown is returned when we attempt to create an image reference
 // for a List that has not yet been saved to an image.
 var ErrListImageUnknown = stderrors.New("unable to determine which image holds the manifest list")
@@ -49,14 +56,18 @@ type List interface {
 // PushOptions includes various settings which are needed for pushing the
 // manifest list and its instances.
 type PushOptions struct {
-	Store              storage.Store
-	SystemContext      *types.SystemContext  // github.com/containers/image/types.SystemContext
-	ImageListSelection cp.ImageListSelection // set to either CopySystemImage, CopyAllImages, or CopySpecificImages
-	Instances          []digest.Digest       // instances to copy if ImageListSelection == CopySpecificImages
-	ReportWriter       io.Writer             // will be used to log the writing of the list and any blobs
-	SignBy             string                // fingerprint of GPG key to use to sign images
-	RemoveSignatures   bool                  // true to discard signatures in images
-	ManifestType       string                // the format to use when saving the list - possible options are oci, v2s1, and v2s2
+	Store                            storage.Store
+	SystemContext                    *types.SystemContext  // github.com/containers/image/types.SystemContext
+	ImageListSelection               cp.ImageListSelection // set to either CopySystemImage, CopyAllImages, or CopySpecificImages
+	Instances                        []digest.Digest       // instances to copy if ImageListSelection == CopySpecificImages
+	ReportWriter                     io.Writer             // will be used to log the writing of the list and any blobs
+	SignBy                           string                // fingerprint of GPG key to use to sign images
+	SignPassphrase                   string                // passphrase to use when signing with the key ID from SignBy.
+	SignBySigstorePrivateKeyFile     string                // if non-empty, asks for a signature to be added during the copy, using a sigstore private key file at the provided path.
+	SignSigstorePrivateKeyPassphrase []byte                // passphrase to use when signing with SignBySigstorePrivateKeyFile.
+	RemoveSignatures                 bool                  // true to discard signatures in images
+	ManifestType                     string                // the format to use when saving the list - possible options are oci, v2s1, and v2s2
+	SourceFilter                     LookupReferenceFunc   // filter the list source
 }
 
 // Create creates a new list containing information about the specified image,
@@ -76,11 +87,11 @@ func Create() List {
 func LoadFromImage(store storage.Store, image string) (string, List, error) {
 	img, err := store.Image(image)
 	if err != nil {
-		return "", nil, errors.Wrapf(err, "error locating image %q for loading manifest list", image)
+		return "", nil, fmt.Errorf("locating image %q for loading manifest list: %w", image, err)
 	}
 	manifestBytes, err := store.ImageBigData(img.ID, storage.ImageDigestManifestBigDataNamePrefix)
 	if err != nil {
-		return "", nil, errors.Wrapf(err, "error locating image %q for loading manifest list", image)
+		return "", nil, fmt.Errorf("locating image %q for loading manifest list: %w", image, err)
 	}
 	manifestList, err := manifests.FromBlob(manifestBytes)
 	if err != nil {
@@ -92,10 +103,10 @@ func LoadFromImage(store storage.Store, image string) (string, List, error) {
 	}
 	instancesBytes, err := store.ImageBigData(img.ID, instancesData)
 	if err != nil {
-		return "", nil, errors.Wrapf(err, "error locating image %q for loading instance list", image)
+		return "", nil, fmt.Errorf("locating image %q for loading instance list: %w", image, err)
 	}
 	if err := json.Unmarshal(instancesBytes, &list.instances); err != nil {
-		return "", nil, errors.Wrapf(err, "error decoding instance list for image %q", image)
+		return "", nil, fmt.Errorf("decoding instance list for image %q: %w", image, err)
 	}
 	list.instances[""] = img.ID
 	return img.ID, list, err
@@ -115,7 +126,7 @@ func (l *list) SaveToImage(store storage.Store, imageID string, names []string,
 		return "", err
 	}
 	img, err := store.CreateImage(imageID, names, "", "", &storage.ImageOptions{})
-	if err == nil || errors.Cause(err) == storage.ErrDuplicateID {
+	if err == nil || errors.Is(err, storage.ErrDuplicateID) {
 		created := (err == nil)
 		if created {
 			imageID = img.ID
@@ -125,34 +136,34 @@ func (l *list) SaveToImage(store storage.Store, imageID string, names []string,
 		if err != nil {
 			if created {
 				if _, err2 := store.DeleteImage(img.ID, true); err2 != nil {
-					logrus.Errorf("error deleting image %q after failing to save manifest for it", img.ID)
+					logrus.Errorf("Deleting image %q after failing to save manifest for it", img.ID)
 				}
 			}
-			return "", errors.Wrapf(err, "error saving manifest list to image %q", imageID)
+			return "", fmt.Errorf("saving manifest list to image %q: %w", imageID, err)
 		}
 		err = store.SetImageBigData(imageID, instancesData, instancesBytes, nil)
 		if err != nil {
 			if created {
 				if _, err2 := store.DeleteImage(img.ID, true); err2 != nil {
-					logrus.Errorf("error deleting image %q after failing to save instance locations for it", img.ID)
+					logrus.Errorf("Deleting image %q after failing to save instance locations for it", img.ID)
 				}
 			}
-			return "", errors.Wrapf(err, "error saving instance list to image %q", imageID)
+			return "", fmt.Errorf("saving instance list to image %q: %w", imageID, err)
 		}
 		return imageID, nil
 	}
-	return "", errors.Wrapf(err, "error creating image to hold manifest list")
+	return "", fmt.Errorf("creating image to hold manifest list: %w", err)
 }
 
 // Reference returns an image reference for the composite image being built
 // in the list, or an error if the list has never been saved to a local image.
 func (l *list) Reference(store storage.Store, multiple cp.ImageListSelection, instances []digest.Digest) (types.ImageReference, error) {
 	if l.instances[""] == "" {
-		return nil, errors.Wrap(ErrListImageUnknown, "error building reference to list")
+		return nil, fmt.Errorf("building reference to list: %w", ErrListImageUnknown)
 	}
 	s, err := is.Transport.ParseStoreReference(store, l.instances[""])
 	if err != nil {
-		return nil, errors.Wrapf(err, "error creating ImageReference from image %q", l.instances[""])
+		return nil, fmt.Errorf("creating ImageReference from image %q: %w", l.instances[""], err)
 	}
 	references := make([]types.ImageReference, 0, len(l.instances))
 	whichInstances := make([]digest.Digest, 0, len(l.instances))
@@ -176,7 +187,7 @@ func (l *list) Reference(store storage.Store, multiple cp.ImageListSelection, in
 		imageName := l.instances[instance]
 		ref, err := alltransports.ParseImageName(imageName)
 		if err != nil {
-			return nil, errors.Wrapf(err, "error creating ImageReference from image %q", imageName)
+			return nil, fmt.Errorf("creating ImageReference from image %q: %w", imageName, err)
 		}
 		references = append(references, ref)
 	}
@@ -188,7 +199,7 @@ func (l *list) Push(ctx context.Context, dest types.ImageReference, options Push
 	// Load the system signing policy.
 	pushPolicy, err := signature.DefaultPolicy(options.SystemContext)
 	if err != nil {
-		return nil, "", errors.Wrapf(err, "error obtaining default signature policy")
+		return nil, "", fmt.Errorf("obtaining default signature policy: %w", err)
 	}
 
 	// Override the settings for local storage to make sure that we can always read the source "image".
@@ -196,11 +207,11 @@ func (l *list) Push(ctx context.Context, dest types.ImageReference, options Push
 
 	policyContext, err := signature.NewPolicyContext(pushPolicy)
 	if err != nil {
-		return nil, "", errors.Wrapf(err, "error creating new signature policy context")
+		return nil, "", fmt.Errorf("creating new signature policy context: %w", err)
 	}
 	defer func() {
 		if err2 := policyContext.Destroy(); err2 != nil {
-			logrus.Errorf("error destroying signature policy context: %v", err2)
+			logrus.Errorf("Destroying signature policy context: %v", err2)
 		}
 	}()
 
@@ -221,15 +232,23 @@ func (l *list) Push(ctx context.Context, dest types.ImageReference, options Push
 	if err != nil {
 		return nil, "", err
 	}
+	if options.SourceFilter != nil {
+		if src, err = options.SourceFilter(src); err != nil {
+			return nil, "", err
+		}
+	}
 	copyOptions := &cp.Options{
-		ImageListSelection:    options.ImageListSelection,
-		Instances:             options.Instances,
-		SourceCtx:             options.SystemContext,
-		DestinationCtx:        options.SystemContext,
-		ReportWriter:          options.ReportWriter,
-		RemoveSignatures:      options.RemoveSignatures,
-		SignBy:                options.SignBy,
-		ForceManifestMIMEType: singleImageManifestType,
+		ImageListSelection:               options.ImageListSelection,
+		Instances:                        options.Instances,
+		SourceCtx:                        options.SystemContext,
+		DestinationCtx:                   options.SystemContext,
+		ReportWriter:                     options.ReportWriter,
+		RemoveSignatures:                 options.RemoveSignatures,
+		SignBy:                           options.SignBy,
+		SignPassphrase:                   options.SignPassphrase,
+		SignBySigstorePrivateKeyFile:     options.SignBySigstorePrivateKeyFile,
+		SignSigstorePrivateKeyPassphrase: options.SignSigstorePrivateKeyPassphrase,
+		ForceManifestMIMEType:            singleImageManifestType,
 	}
 
 	// Copy whatever we were asked to copy.
@@ -254,7 +273,7 @@ func (l *list) Push(ctx context.Context, dest types.ImageReference, options Push
 func (l *list) Add(ctx context.Context, sys *types.SystemContext, ref types.ImageReference, all bool) (digest.Digest, error) {
 	src, err := ref.NewImageSource(ctx, sys)
 	if err != nil {
-		return "", errors.Wrapf(err, "error setting up to read manifest and configuration from %q", transports.ImageName(ref))
+		return "", fmt.Errorf("setting up to read manifest and configuration from %q: %w", transports.ImageName(ref), err)
 	}
 	defer src.Close()
 
@@ -269,13 +288,13 @@ func (l *list) Add(ctx context.Context, sys *types.SystemContext, ref types.Imag
 
 	primaryManifestBytes, primaryManifestType, err := src.GetManifest(ctx, nil)
 	if err != nil {
-		return "", errors.Wrapf(err, "error reading manifest from %q", transports.ImageName(ref))
+		return "", fmt.Errorf("reading manifest from %q: %w", transports.ImageName(ref), err)
 	}
 
 	if manifest.MIMETypeIsMultiImage(primaryManifestType) {
 		lists, err := manifests.FromBlob(primaryManifestBytes)
 		if err != nil {
-			return "", errors.Wrapf(err, "error parsing manifest list in %q", transports.ImageName(ref))
+			return "", fmt.Errorf("parsing manifest list in %q: %w", transports.ImageName(ref), err)
 		}
 		if all {
 			for i, instance := range lists.OCIv1().Manifests {
@@ -299,11 +318,11 @@ func (l *list) Add(ctx context.Context, sys *types.SystemContext, ref types.Imag
 		} else {
 			list, err := manifest.ListFromBlob(primaryManifestBytes, primaryManifestType)
 			if err != nil {
-				return "", errors.Wrapf(err, "error parsing manifest list in %q", transports.ImageName(ref))
+				return "", fmt.Errorf("parsing manifest list in %q: %w", transports.ImageName(ref), err)
 			}
 			instanceDigest, err := list.ChooseInstance(sys)
 			if err != nil {
-				return "", errors.Wrapf(err, "error selecting image from manifest list in %q", transports.ImageName(ref))
+				return "", fmt.Errorf("selecting image from manifest list in %q: %w", transports.ImageName(ref), err)
 			}
 			added := false
 			for i, instance := range lists.OCIv1().Manifests {
@@ -345,38 +364,39 @@ func (l *list) Add(ctx context.Context, sys *types.SystemContext, ref types.Imag
 		if instanceInfo.OS == "" || instanceInfo.Architecture == "" {
 			img, err := image.FromUnparsedImage(ctx, sys, image.UnparsedInstance(src, instanceInfo.instanceDigest))
 			if err != nil {
-				return "", errors.Wrapf(err, "error reading configuration blob from %q", transports.ImageName(ref))
+				return "", fmt.Errorf("reading configuration blob from %q: %w", transports.ImageName(ref), err)
 			}
 			config, err := img.OCIConfig(ctx)
 			if err != nil {
-				return "", errors.Wrapf(err, "error reading info about config blob from %q", transports.ImageName(ref))
+				return "", fmt.Errorf("reading info about config blob from %q: %w", transports.ImageName(ref), err)
 			}
 			if instanceInfo.OS == "" {
 				instanceInfo.OS = config.OS
+				instanceInfo.OSVersion = config.OSVersion
+				instanceInfo.OSFeatures = config.OSFeatures
 			}
 			if instanceInfo.Architecture == "" {
 				instanceInfo.Architecture = config.Architecture
+				instanceInfo.Variant = config.Variant
 			}
 		}
 		manifestBytes, manifestType, err := src.GetManifest(ctx, instanceInfo.instanceDigest)
 		if err != nil {
-			return "", errors.Wrapf(err, "error reading manifest from %q, instance %q", transports.ImageName(ref), instanceInfo.instanceDigest)
+			return "", fmt.Errorf("reading manifest from %q, instance %q: %w", transports.ImageName(ref), instanceInfo.instanceDigest, err)
 		}
 		if instanceInfo.instanceDigest == nil {
 			manifestDigest, err = manifest.Digest(manifestBytes)
 			if err != nil {
-				return "", errors.Wrapf(err, "error computing digest of manifest from %q", transports.ImageName(ref))
+				return "", fmt.Errorf("computing digest of manifest from %q: %w", transports.ImageName(ref), err)
 			}
 			instanceInfo.instanceDigest = &manifestDigest
 			instanceInfo.Size = int64(len(manifestBytes))
-		} else {
-			if manifestDigest == "" {
-				manifestDigest = *instanceInfo.instanceDigest
-			}
+		} else if manifestDigest == "" {
+			manifestDigest = *instanceInfo.instanceDigest
 		}
 		err = l.List.AddInstance(*instanceInfo.instanceDigest, instanceInfo.Size, manifestType, instanceInfo.OS, instanceInfo.Architecture, instanceInfo.OSVersion, instanceInfo.OSFeatures, instanceInfo.Variant, instanceInfo.Features, instanceInfo.Annotations)
 		if err != nil {
-			return "", errors.Wrapf(err, "error adding instance with digest %q", *instanceInfo.instanceDigest)
+			return "", fmt.Errorf("adding instance with digest %q: %w", *instanceInfo.instanceDigest, err)
 		}
 		if _, ok := l.instances[*instanceInfo.instanceDigest]; !ok {
 			l.instances[*instanceInfo.instanceDigest] = transports.ImageName(ref)
@@ -390,9 +410,7 @@ func (l *list) Add(ctx context.Context, sys *types.SystemContext, ref types.Imag
 func (l *list) Remove(instanceDigest digest.Digest) error {
 	err := l.List.Remove(instanceDigest)
 	if err == nil {
-		if _, needToDelete := l.instances[instanceDigest]; needToDelete {
-			delete(l.instances, instanceDigest)
-		}
+		delete(l.instances, instanceDigest)
 	}
 	return err
 }
@@ -405,11 +423,11 @@ func (l *list) Remove(instanceDigest digest.Digest) error {
 func LockerForImage(store storage.Store, image string) (lockfile.Locker, error) {
 	img, err := store.Image(image)
 	if err != nil {
-		return nil, errors.Wrapf(err, "locating image %q for locating lock", image)
+		return nil, fmt.Errorf("locating image %q for locating lock: %w", image, err)
 	}
 	d := digest.NewDigestFromEncoded(digest.Canonical, img.ID)
 	if err := d.Validate(); err != nil {
-		return nil, errors.Wrapf(err, "coercing image ID for %q into a digest", image)
+		return nil, fmt.Errorf("coercing image ID for %q into a digest: %w", image, err)
 	}
 	return store.GetDigestLock(d)
 }
