@@ -9,7 +9,7 @@ import (
 
 	"github.com/containers/common/pkg/apparmor"
 	"github.com/containers/common/pkg/capabilities"
-	. "github.com/onsi/ginkgo"
+	. "github.com/onsi/ginkgo/v2"
 	"github.com/onsi/gomega"
 	selinux "github.com/opencontainers/selinux/go-selinux"
 	"github.com/sirupsen/logrus"
@@ -27,12 +27,23 @@ var _ = Describe("Config", func() {
 			// Then
 			gomega.Expect(err).To(gomega.BeNil())
 			gomega.Expect(defaultConfig.Containers.ApparmorProfile).To(gomega.Equal(apparmor.Profile))
+			gomega.Expect(defaultConfig.Containers.BaseHostsFile).To(gomega.Equal(""))
 			gomega.Expect(defaultConfig.Containers.PidsLimit).To(gomega.BeEquivalentTo(2048))
+			gomega.Expect(defaultConfig.Engine.ServiceTimeout).To(gomega.BeEquivalentTo(5))
+			gomega.Expect(defaultConfig.NetNS()).To(gomega.BeEquivalentTo("private"))
+			gomega.Expect(defaultConfig.IPCNS()).To(gomega.BeEquivalentTo("shareable"))
+			gomega.Expect(defaultConfig.Engine.InfraImage).To(gomega.BeEquivalentTo(""))
+			gomega.Expect(defaultConfig.Engine.ImageVolumeMode).To(gomega.BeEquivalentTo("bind"))
+			gomega.Expect(defaultConfig.Engine.SSHConfig).To(gomega.ContainSubstring("/.ssh/config"))
+			path, err := defaultConfig.ImageCopyTmpDir()
+			gomega.Expect(err).To(gomega.BeNil())
+			gomega.Expect(path).To(gomega.BeEquivalentTo("/var/tmp"))
 		})
 
 		It("should succeed with devices", func() {
 			// Given
-			sut.Containers.Devices = []string{"/dev/null:/dev/null:rw",
+			sut.Containers.Devices = []string{
+				"/dev/null:/dev/null:rw",
 				"/dev/sdc/",
 				"/dev/sdc:/dev/xvdc",
 				"/dev/sdc:rm",
@@ -91,9 +102,7 @@ var _ = Describe("Config", func() {
 			defaultConfig, _ := NewConfig("")
 			// EnableLabeling should match whether or not SELinux is enabled on the host
 			gomega.Expect(defaultConfig.Containers.EnableLabeling).To(gomega.Equal(selinux.GetEnabled()))
-
 		})
-
 	})
 
 	Describe("ValidateNetworkConfig", func() {
@@ -107,6 +116,35 @@ var _ = Describe("Config", func() {
 		})
 	})
 
+	Describe("readStorageTmp", func() {
+		It("test image_copy_tmp_dir='storage'", func() {
+			// Reload from new configuration file
+			testFile := "testdata/temp.conf"
+			content := `[engine]
+image_copy_tmp_dir="storage"`
+			err := ioutil.WriteFile(testFile, []byte(content), os.ModePerm)
+			// Then
+			gomega.Expect(err).To(gomega.BeNil())
+			defer os.Remove(testFile)
+
+			config, _ := NewConfig(testFile)
+			path, err := config.ImageCopyTmpDir()
+			gomega.Expect(err).To(gomega.BeNil())
+			gomega.Expect(path).To(gomega.ContainSubstring("containers/storage/tmp"))
+			// Given we do
+			oldTMPDIR, set := os.LookupEnv("TMPDIR")
+			os.Setenv("TMPDIR", "/var/tmp/foobar")
+			path, err = config.ImageCopyTmpDir()
+			gomega.Expect(err).To(gomega.BeNil())
+			gomega.Expect(path).To(gomega.BeEquivalentTo("/var/tmp/foobar"))
+			if set {
+				os.Setenv("TMPDIR", oldTMPDIR)
+			} else {
+				os.Unsetenv("TMPDIR")
+			}
+		})
+	})
+
 	Describe("readConfigFromFile", func() {
 		It("should succeed with default config", func() {
 			// Given
@@ -119,7 +157,6 @@ var _ = Describe("Config", func() {
 
 			OCIRuntimeMap := map[string][]string{
 				"kata": {
-
 					"/usr/bin/kata-runtime",
 					"/usr/sbin/kata-runtime",
 					"/usr/local/bin/kata-runtime",
@@ -138,6 +175,9 @@ var _ = Describe("Config", func() {
 					"/bin/runc",
 					"/usr/lib/cri-o-runc/sbin/runc",
 				},
+				"runj": {
+					"/usr/local/bin/runj",
+				},
 				"crun": {
 					"/usr/bin/crun",
 					"/usr/local/bin/crun",
@@ -151,11 +191,15 @@ var _ = Describe("Config", func() {
 					"/sbin/runsc",
 					"/run/current-system/sw/bin/runsc",
 				},
+				"krun": {
+					"/usr/bin/krun",
+					"/usr/local/bin/krun",
+				},
 			}
 
 			pluginDirs := []string{
 				"/usr/libexec/cni",
-				"/usr/libexec/foo",
+				"/tmp",
 			}
 
 			envs := []string{
@@ -163,6 +207,14 @@ var _ = Describe("Config", func() {
 				"TERM=xterm",
 			}
 
+			volumes := []string{
+				"$HOME:$HOME",
+			}
+
+			newVolumes := []string{
+				os.ExpandEnv("$HOME:$HOME"),
+			}
+
 			helperDirs := []string{
 				"/somepath",
 			}
@@ -176,11 +228,22 @@ var _ = Describe("Config", func() {
 			gomega.Expect(defaultConfig.Engine.NumLocks).To(gomega.BeEquivalentTo(2048))
 			gomega.Expect(defaultConfig.Engine.OCIRuntimes).To(gomega.Equal(OCIRuntimeMap))
 			gomega.Expect(defaultConfig.Containers.HTTPProxy).To(gomega.Equal(false))
+			gomega.Expect(defaultConfig.Engine.NetworkCmdOptions).To(gomega.BeNil())
 			gomega.Expect(defaultConfig.Engine.HelperBinariesDir).To(gomega.Equal(helperDirs))
+			gomega.Expect(defaultConfig.Engine.ServiceTimeout).To(gomega.BeEquivalentTo(300))
+			gomega.Expect(defaultConfig.Engine.InfraImage).To(gomega.BeEquivalentTo("k8s.gcr.io/pause:3.4.1"))
+			gomega.Expect(defaultConfig.Machine.Volumes).To(gomega.BeEquivalentTo(volumes))
+			newV, err := defaultConfig.MachineVolumes()
+			if newVolumes[0] == ":" {
+				// $HOME is not set
+				gomega.Expect(err).To(gomega.HaveOccurred())
+			} else {
+				gomega.Expect(err).To(gomega.BeNil())
+				gomega.Expect(newV).To(gomega.BeEquivalentTo(newVolumes))
+			}
 		})
 
 		It("test GetDefaultEnvEx", func() {
-
 			envs := []string{
 				"PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin",
 				"TERM=xterm",
@@ -286,19 +349,25 @@ var _ = Describe("Config", func() {
 			gomega.Expect(config.Containers.ApparmorProfile).To(gomega.Equal(apparmor.Profile))
 			gomega.Expect(config.Containers.PidsLimit).To(gomega.BeEquivalentTo(2048))
 			gomega.Expect(config.Containers.Env).To(gomega.BeEquivalentTo(envs))
+			gomega.Expect(config.Containers.UserNS).To(gomega.BeEquivalentTo(""))
 			gomega.Expect(config.Network.CNIPluginDirs).To(gomega.Equal(DefaultCNIPluginDirs))
 			gomega.Expect(config.Engine.NumLocks).To(gomega.BeEquivalentTo(2048))
 			gomega.Expect(config.Engine.OCIRuntimes["runc"]).To(gomega.Equal(OCIRuntimeMap["runc"]))
 			if useSystemd() {
 				gomega.Expect(config.Engine.CgroupManager).To(gomega.BeEquivalentTo("systemd"))
-				gomega.Expect(config.Engine.EventsLogger).To(gomega.BeEquivalentTo("journald"))
-				gomega.Expect(config.Containers.LogDriver).To(gomega.BeEquivalentTo("k8s-file"))
 			} else {
 				gomega.Expect(config.Engine.CgroupManager).To(gomega.BeEquivalentTo("cgroupfs"))
+			}
+			if useJournald() {
+				gomega.Expect(config.Engine.EventsLogger).To(gomega.BeEquivalentTo("journald"))
+				gomega.Expect(config.Containers.LogDriver).To(gomega.BeEquivalentTo("journald"))
+			} else {
 				gomega.Expect(config.Engine.EventsLogger).To(gomega.BeEquivalentTo("file"))
 				gomega.Expect(config.Containers.LogDriver).To(gomega.BeEquivalentTo("k8s-file"))
 			}
-
+			gomega.Expect(config.Engine.EventsLogFilePath).To(gomega.BeEquivalentTo(""))
+			gomega.Expect(uint64(config.Engine.EventsLogFileMaxSize)).To(gomega.Equal(DefaultEventsLogSizeMax))
+			gomega.Expect(config.Engine.PodExitPolicy).To(gomega.Equal(PodExitPolicyContinue))
 		})
 
 		It("should success with valid user file path", func() {
@@ -309,6 +378,10 @@ var _ = Describe("Config", func() {
 			gomega.Expect(err).To(gomega.BeNil())
 			gomega.Expect(config.Containers.ApparmorProfile).To(gomega.Equal("container-default"))
 			gomega.Expect(config.Containers.PidsLimit).To(gomega.BeEquivalentTo(2048))
+			gomega.Expect(config.Containers.BaseHostsFile).To(gomega.BeEquivalentTo("/etc/hosts2"))
+			gomega.Expect(config.Containers.HostContainersInternalIP).To(gomega.BeEquivalentTo("1.2.3.4"))
+			gomega.Expect(config.Engine.ImageVolumeMode).To(gomega.BeEquivalentTo("tmpfs"))
+			gomega.Expect(config.Engine.SSHConfig).To(gomega.Equal("/foo/bar/.ssh/config"))
 		})
 
 		It("contents of passed-in file should override others", func() {
@@ -332,6 +405,12 @@ var _ = Describe("Config", func() {
 			gomega.Expect(config.Containers.LogSizeMax).To(gomega.Equal(int64(100000)))
 			gomega.Expect(config.Engine.ImageParallelCopies).To(gomega.Equal(uint(10)))
 			gomega.Expect(config.Engine.ImageDefaultFormat).To(gomega.Equal("v2s2"))
+			gomega.Expect(config.Engine.EventsLogFilePath).To(gomega.BeEquivalentTo("/tmp/events.log"))
+			path, err := config.ImageCopyTmpDir()
+			gomega.Expect(err).To(gomega.BeNil())
+			gomega.Expect(path).To(gomega.BeEquivalentTo("/tmp/foobar"))
+			gomega.Expect(uint64(config.Engine.EventsLogFileMaxSize)).To(gomega.Equal(uint64(500)))
+			gomega.Expect(config.Engine.PodExitPolicy).To(gomega.BeEquivalentTo(PodExitPolicyStop))
 		})
 
 		It("should fail with invalid value", func() {
@@ -509,9 +588,10 @@ var _ = Describe("Config", func() {
 			cfg, err = ReadCustomConfig()
 			gomega.Expect(err).ShouldNot(gomega.HaveOccurred())
 
-			u, i, err := cfg.ActiveDestination()
+			u, i, m, err := cfg.ActiveDestination()
 			gomega.Expect(err).ShouldNot(gomega.HaveOccurred())
 
+			gomega.Expect(m).To(gomega.Equal(false))
 			gomega.Expect(u).To(gomega.Equal("https://qa/run/podman/podman.sock"))
 			gomega.Expect(i).To(gomega.Equal("/.ssh/id_rsa"))
 		})
@@ -541,7 +621,7 @@ var _ = Describe("Config", func() {
 			oldContainerConnection, hostEnvSet := os.LookupEnv("CONTAINER_CONNECTION")
 			os.Setenv("CONTAINER_CONNECTION", "QB")
 
-			u, i, err := cfg.ActiveDestination()
+			u, i, m, err := cfg.ActiveDestination()
 			// Undo that
 			if hostEnvSet {
 				os.Setenv("CONTAINER_CONNECTION", oldContainerConnection)
@@ -549,6 +629,7 @@ var _ = Describe("Config", func() {
 				os.Unsetenv("CONTAINER_CONNECTION")
 			}
 			gomega.Expect(err).ShouldNot(gomega.HaveOccurred())
+			gomega.Expect(m).To(gomega.Equal(false))
 
 			gomega.Expect(u).To(gomega.Equal("https://qb/run/podman/podman.sock"))
 			gomega.Expect(i).To(gomega.Equal("/.ssh/qb_id_rsa"))
@@ -581,7 +662,8 @@ var _ = Describe("Config", func() {
 			os.Setenv("CONTAINER_HOST", "foo.bar")
 			os.Setenv("CONTAINER_SSHKEY", "/.ssh/newid_rsa")
 
-			u, i, err := cfg.ActiveDestination()
+			u, i, m, err := cfg.ActiveDestination()
+
 			// Undo that
 			if hostEnvSet {
 				os.Setenv("CONTAINER_HOST", oldContainerHost)
@@ -596,6 +678,7 @@ var _ = Describe("Config", func() {
 			}
 
 			gomega.Expect(err).ShouldNot(gomega.HaveOccurred())
+			gomega.Expect(m).To(gomega.Equal(false))
 
 			gomega.Expect(u).To(gomega.Equal("foo.bar"))
 			gomega.Expect(i).To(gomega.Equal("/.ssh/newid_rsa"))
@@ -605,7 +688,7 @@ var _ = Describe("Config", func() {
 			cfg, err := ReadCustomConfig()
 			gomega.Expect(err).ShouldNot(gomega.HaveOccurred())
 
-			_, _, err = cfg.ActiveDestination()
+			_, _, _, err = cfg.ActiveDestination()
 			gomega.Expect(err).Should(gomega.HaveOccurred())
 		})
 
@@ -615,9 +698,9 @@ var _ = Describe("Config", func() {
 				gomega.Expect(err).ShouldNot(gomega.HaveOccurred())
 				conf := file.Name() + ".conf"
 
-				os.Rename(file.Name(), conf)
+				err = os.Rename(file.Name(), conf)
+				gomega.Expect(err).ShouldNot(gomega.HaveOccurred())
 				return conf
-
 			}
 			configs := []string{
 				"test1",
@@ -655,6 +738,7 @@ var _ = Describe("Config", func() {
 			conf := Config{}
 			content := bytes.NewBufferString("")
 			logrus.SetOutput(content)
+			logrus.SetLevel(logrus.DebugLevel)
 			err := readConfigFromFile("testdata/containers_broken.conf", &conf)
 			gomega.Expect(err).To(gomega.BeNil())
 			gomega.Expect(conf.Containers.NetNS).To(gomega.Equal("bridge"))
@@ -716,4 +800,33 @@ env=["foo=bar"]`
 			gomega.Expect(err).To(gomega.BeNil())
 		})
 	})
+
+	It("write default config should be empty", func() {
+		defer os.Unsetenv("CONTAINERS_CONF")
+		os.Setenv("CONTAINERS_CONF", "/dev/null")
+		conf, err := ReadCustomConfig()
+		gomega.Expect(err).ToNot(gomega.HaveOccurred())
+
+		f, err := ioutil.TempFile("", "container-common-test")
+		gomega.Expect(err).ToNot(gomega.HaveOccurred())
+		defer f.Close()
+		defer os.Remove(f.Name())
+		os.Setenv("CONTAINERS_CONF", f.Name())
+		err = conf.Write()
+		gomega.Expect(err).ToNot(gomega.HaveOccurred())
+		b, err := ioutil.ReadFile(f.Name())
+		gomega.Expect(err).ToNot(gomega.HaveOccurred())
+		// config should only contain empty stanzas
+		gomega.Expect(string(b)).To(gomega.
+			Equal("[containers]\n\n[engine]\n\n[machine]\n\n[network]\n\n[secrets]\n\n[configmaps]\n"))
+	})
+
+	It("validate ImageVolumeMode", func() {
+		for _, mode := range append(validImageVolumeModes, "") {
+			err := ValidateImageVolumeMode(mode)
+			gomega.Expect(err).ToNot(gomega.HaveOccurred())
+		}
+		err := ValidateImageVolumeMode("bogus")
+		gomega.Expect(err).To(gomega.HaveOccurred())
+	})
 })
