@@ -32,46 +32,73 @@ func TestAddSecretAndLookupData(t *testing.T) {
 	require.NoError(t, err)
 	defer cleanup(testpath)
 
-	_, err = manager.Store("mysecret", []byte("mydata"), drivertype, opts)
+	metaData := make(map[string]string)
+	metaData["immutable"] = "true"
+
+	labels := make(map[string]string)
+	labels["foo"] = "bar"
+	labels["another"] = "label"
+
+	storeOpts := StoreOptions{
+		DriverOpts: opts,
+		Metadata:   metaData,
+		Labels:     labels,
+	}
+
+	_, err = manager.Store("mysecret", []byte("mydata"), drivertype, storeOpts)
 	require.NoError(t, err)
 
 	_, err = manager.lookupSecret("mysecret")
 	require.NoError(t, err)
 
-	_, data, err := manager.LookupSecretData("mysecret")
+	s, data, err := manager.LookupSecretData("mysecret")
 	require.NoError(t, err)
 	if !bytes.Equal(data, []byte("mydata")) {
 		t.Errorf("error: secret data not equal")
 	}
+	if val, ok := s.Metadata["immutable"]; !ok || val != "true" {
+		t.Errorf("error: no metadata")
+	}
+	if val, ok := s.Labels["foo"]; !ok || val != "bar" {
+		t.Errorf("error: label incorrect")
+	}
+	if len(s.Labels) != 2 {
+		t.Errorf("error: incorrect number of labels")
+	}
 }
+
 func TestAddSecretName(t *testing.T) {
 	manager, testpath, err := setup()
 	require.NoError(t, err)
 	defer cleanup(testpath)
 
+	storeOpts := StoreOptions{
+		DriverOpts: opts,
+	}
+
 	// test one char secret name
-	_, err = manager.Store("a", []byte("mydata"), drivertype, opts)
+	_, err = manager.Store("a", []byte("mydata"), drivertype, storeOpts)
 	require.NoError(t, err)
 
 	_, err = manager.lookupSecret("a")
 	require.NoError(t, err)
 
 	// name too short
-	_, err = manager.Store("", []byte("mydata"), drivertype, opts)
+	_, err = manager.Store("", []byte("mydata"), drivertype, storeOpts)
 	require.Error(t, err)
 	// name too long
-	_, err = manager.Store("uatqsbssrapurkuqoapubpifvsrissslzjehalxcesbhpxcvhsozlptrmngrivaiz", []byte("mydata"), drivertype, opts)
+	_, err = manager.Store("uatqsbssrapurkuqoapubpifvsrissslzjehalxcesbhpxcvhsozlptrmngrivaiz", []byte("mydata"), drivertype, storeOpts)
 	require.Error(t, err)
 	// invalid chars
-	_, err = manager.Store("??", []byte("mydata"), drivertype, opts)
+	_, err = manager.Store("??", []byte("mydata"), drivertype, storeOpts)
 	require.Error(t, err)
-	_, err = manager.Store("-a", []byte("mydata"), drivertype, opts)
+	_, err = manager.Store("-a", []byte("mydata"), drivertype, storeOpts)
 	require.Error(t, err)
-	_, err = manager.Store("a-", []byte("mydata"), drivertype, opts)
+	_, err = manager.Store("a-", []byte("mydata"), drivertype, storeOpts)
 	require.Error(t, err)
-	_, err = manager.Store(".a", []byte("mydata"), drivertype, opts)
+	_, err = manager.Store(".a", []byte("mydata"), drivertype, storeOpts)
 	require.Error(t, err)
-	_, err = manager.Store("a.", []byte("mydata"), drivertype, opts)
+	_, err = manager.Store("a.", []byte("mydata"), drivertype, storeOpts)
 	require.Error(t, err)
 }
 
@@ -80,10 +107,14 @@ func TestAddMultipleSecrets(t *testing.T) {
 	require.NoError(t, err)
 	defer cleanup(testpath)
 
-	id, err := manager.Store("mysecret", []byte("mydata"), drivertype, opts)
+	storeOpts := StoreOptions{
+		DriverOpts: opts,
+	}
+
+	id, err := manager.Store("mysecret", []byte("mydata"), drivertype, storeOpts)
 	require.NoError(t, err)
 
-	id2, err := manager.Store("mysecret2", []byte("mydata2"), drivertype, opts)
+	id2, err := manager.Store("mysecret2", []byte("mydata2"), drivertype, storeOpts)
 	require.NoError(t, err)
 
 	secrets, err := manager.List()
@@ -114,10 +145,14 @@ func TestAddSecretDupName(t *testing.T) {
 	require.NoError(t, err)
 	defer cleanup(testpath)
 
-	_, err = manager.Store("mysecret", []byte("mydata"), drivertype, opts)
+	storeOpts := StoreOptions{
+		DriverOpts: opts,
+	}
+
+	_, err = manager.Store("mysecret", []byte("mydata"), drivertype, storeOpts)
 	require.NoError(t, err)
 
-	_, err = manager.Store("mysecret", []byte("mydata"), drivertype, opts)
+	_, err = manager.Store("mysecret", []byte("mydata"), drivertype, storeOpts)
 	require.Error(t, err)
 }
 
@@ -126,12 +161,16 @@ func TestAddSecretPrefix(t *testing.T) {
 	require.NoError(t, err)
 	defer cleanup(testpath)
 
+	storeOpts := StoreOptions{
+		DriverOpts: opts,
+	}
+
 	// If the randomly generated secret id is something like "abcdeiuoergnadufigh"
 	// we should still allow someone to store a secret with the name "abcd" or "a"
-	secretID, err := manager.Store("mysecret", []byte("mydata"), drivertype, opts)
+	secretID, err := manager.Store("mysecret", []byte("mydata"), drivertype, storeOpts)
 	require.NoError(t, err)
 
-	_, err = manager.Store(secretID[0:5], []byte("mydata"), drivertype, opts)
+	_, err = manager.Store(secretID[0:5], []byte("mydata"), drivertype, storeOpts)
 	require.NoError(t, err)
 }
 
@@ -140,7 +179,11 @@ func TestRemoveSecret(t *testing.T) {
 	require.NoError(t, err)
 	defer cleanup(testpath)
 
-	_, err = manager.Store("mysecret", []byte("mydata"), drivertype, opts)
+	storeOpts := StoreOptions{
+		DriverOpts: opts,
+	}
+
+	_, err = manager.Store("mysecret", []byte("mydata"), drivertype, storeOpts)
 	require.NoError(t, err)
 
 	_, err = manager.lookupSecret("mysecret")
@@ -170,7 +213,11 @@ func TestLookupAllSecrets(t *testing.T) {
 	require.NoError(t, err)
 	defer cleanup(testpath)
 
-	id, err := manager.Store("mysecret", []byte("mydata"), drivertype, opts)
+	storeOpts := StoreOptions{
+		DriverOpts: opts,
+	}
+
+	id, err := manager.Store("mysecret", []byte("mydata"), drivertype, storeOpts)
 	require.NoError(t, err)
 
 	// inspect using secret name
@@ -184,7 +231,11 @@ func TestInspectSecretId(t *testing.T) {
 	require.NoError(t, err)
 	defer cleanup(testpath)
 
-	id, err := manager.Store("mysecret", []byte("mydata"), drivertype, opts)
+	storeOpts := StoreOptions{
+		DriverOpts: opts,
+	}
+
+	id, err := manager.Store("mysecret", []byte("mydata"), drivertype, storeOpts)
 	require.NoError(t, err)
 
 	_, err = manager.lookupSecret("mysecret")
@@ -216,9 +267,13 @@ func TestSecretList(t *testing.T) {
 	require.NoError(t, err)
 	defer cleanup(testpath)
 
-	_, err = manager.Store("mysecret", []byte("mydata"), drivertype, opts)
+	storeOpts := StoreOptions{
+		DriverOpts: opts,
+	}
+
+	_, err = manager.Store("mysecret", []byte("mydata"), drivertype, storeOpts)
 	require.NoError(t, err)
-	_, err = manager.Store("mysecret2", []byte("mydata2"), drivertype, opts)
+	_, err = manager.Store("mysecret2", []byte("mydata2"), drivertype, storeOpts)
 	require.NoError(t, err)
 
 	allSecrets, err := manager.List()
